@@ -42,7 +42,11 @@ bladeRF_SoapySDR::bladeRF_SoapySDR(const bladerf_devinfo &devinfo):
     _rxBuffSize(0),
     _txBuffSize(0),
     _rxMinTimeoutMs(0),
+    _xb200Mode("disabled"),
+    _samplingMode("internal"),
+    _loopbackMode("disabled"),
     _dev(NULL)
+
 {
     bladerf_devinfo info = devinfo;
     SoapySDR::logf(SOAPY_SDR_INFO, "bladerf_open_with_devinfo()");
@@ -130,10 +134,166 @@ std::string bladeRF_SoapySDR::getAntenna(const int direction, const size_t chann
     return SoapySDR::Device::getAntenna(direction, channel);
 }
 
+/*******************************************************************
+ * Calibration API
+ ******************************************************************/
+
+bool bladeRF_SoapySDR::hasDCOffset(const int, const size_t) const
+{
+    return true;
+}
+
+void bladeRF_SoapySDR::setDCOffset(const int direction, const size_t, const std::complex<double> &offset)
+{
+    int ret = 0;
+    int16_t i = 0;
+    int16_t q = 0;
+
+    if (offset.real() > 1.0)
+        i = int16_t(1.0 * 2048);
+    else
+        i = int16_t(offset.real() * 2048);
+
+    if (offset.imag() > 1.0)
+        q = int16_t(1.0 * 2048);
+    else
+        q = int16_t(offset.imag() * 2048);
+
+    ret = bladerf_set_correction(_dev, _dir2mod(direction), BLADERF_CORR_LMS_DCOFF_I, i);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_correction(%f) returned %s", i, _err2str(ret).c_str());
+        throw std::runtime_error("setDCOffset() " + _err2str(ret));
+    }
+
+    ret = bladerf_set_correction(_dev, _dir2mod(direction), BLADERF_CORR_LMS_DCOFF_Q, q);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_correction(%f) returned %s", q, _err2str(ret).c_str());
+        throw std::runtime_error("setDCOffset() " + _err2str(ret));
+    }
+}
+
+std::complex<double> bladeRF_SoapySDR::getDCOffset(const int direction, const size_t) const
+{
+    int ret = 0;
+    int16_t i = 0;
+    int16_t q = 0;
+
+    ret = bladerf_get_correction(_dev, _dir2mod(direction), BLADERF_CORR_LMS_DCOFF_I, &i);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_correction() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getDCOffset() " + _err2str(ret));
+    }
+
+    ret = bladerf_get_correction(_dev, _dir2mod(direction), BLADERF_CORR_LMS_DCOFF_Q, &q);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_correction() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getDCOffset() " + _err2str(ret));
+    }
+
+    std::complex<double> z(i / 2048.0f, q / 2048.0f);
+    return z;
+}
+
+bool bladeRF_SoapySDR::hasIQBalance(const int, const size_t) const
+{
+    return true;
+}
+
+void bladeRF_SoapySDR::setIQBalance(const int direction, const size_t, const std::complex<double> &balance)
+{
+    int ret = 0;
+    int16_t gain = 0;
+    int16_t phase = 0;
+
+    if (balance.real() > 1.0)
+        gain = int16_t(1.0 * 4096);
+    else
+        gain = int16_t(balance.real() * 4096);
+
+    if (balance.imag() > 1.0)
+        phase = int16_t(1.0 * 4096);
+    else
+        phase = int16_t(balance.imag() * 4096);
+
+    ret = bladerf_set_correction(_dev, _dir2mod(direction), BLADERF_CORR_FPGA_GAIN, gain);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_correction(%f) returned %s", gain, _err2str(ret).c_str());
+        throw std::runtime_error("setIQBalance() " + _err2str(ret));
+    }
+
+    ret = bladerf_set_correction(_dev, _dir2mod(direction), BLADERF_CORR_FPGA_PHASE, phase);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_correction(%f) returned %s", phase, _err2str(ret).c_str());
+        throw std::runtime_error("setIQBalance() " + _err2str(ret));
+    }
+}
+
+std::complex<double> bladeRF_SoapySDR::getIQBalance(const int direction, const size_t) const
+{
+    int ret = 0;
+    int16_t gain = 0;
+    int16_t phase = 0;
+
+    ret = bladerf_get_correction(_dev, _dir2mod(direction), BLADERF_CORR_FPGA_GAIN, &gain);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_correction() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getIQBalance() " + _err2str(ret));
+    }
+
+    ret = bladerf_get_correction(_dev, _dir2mod(direction), BLADERF_CORR_FPGA_PHASE, &phase);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_correction() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getIQBalance() " + _err2str(ret));
+    }
+
+    std::complex<double> z(gain / 4096.0f, phase / 4096.0f);
+    return z;
+}
+
 /*******************************************************************
  * Gain API
  ******************************************************************/
 
+bool bladeRF_SoapySDR::hasGainMode(const int direction, const size_t) const
+{
+    return _dir2mod(direction) == BLADERF_MODULE_RX ? true : false;
+}
+
+void bladeRF_SoapySDR::setGainMode(const int direction, const size_t, const bool automatic)
+{
+    bladerf_gain_mode gain_mode = automatic ? BLADERF_GAIN_AUTOMATIC : BLADERF_GAIN_MANUAL;
+    const int ret = bladerf_set_gain_mode(_dev, _dir2mod(direction), gain_mode);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_set_gain_mode(%f) returned %s", gain_mode, _err2str(ret).c_str());
+        throw std::runtime_error("setGainMode() " + _err2str(ret));
+    }
+}
+
+bool bladeRF_SoapySDR::getGainMode(const int direction, const size_t) const
+{
+    int ret = 0;
+    bladerf_gain_mode gain_mode;
+    bool automatic;
+    ret = bladerf_get_gain_mode(_dev, _dir2mod(direction), &gain_mode);
+    if (ret != 0)
+    {
+        SoapySDR::logf(SOAPY_SDR_ERROR, "bladerf_get_gain_mode() returned %s", _err2str(ret).c_str());
+        throw std::runtime_error("getGainMode() " + _err2str(ret));
+    }
+
+    automatic = gain_mode == BLADERF_GAIN_AUTOMATIC ? true : false;
+    return automatic;
+}
+
 std::vector<std::string> bladeRF_SoapySDR::listGains(const int direction, const size_t) const
 {
     std::vector<std::string> options;
@@ -534,6 +694,20 @@ SoapySDR::ArgInfoList bladeRF_SoapySDR::getSettingInfo(void) const
     return setArgs;
 }
 
+std::string bladeRF_SoapySDR::readSetting(const std::string &key) const
+{
+    if (key == "xb200") {
+        return _xb200Mode;
+    } else if (key == "sampling_mode") {
+        return _samplingMode;
+    } else if (key == "loopback") {
+        return _loopbackMode;
+    }
+
+    SoapySDR_logf(SOAPY_SDR_WARNING, "Unknown setting '%s'", key.c_str());
+    return "";
+}
+
 void bladeRF_SoapySDR::writeSetting(const std::string &key, const std::string &value)
 {
     if (key == "xb200")
@@ -543,6 +717,7 @@ void bladeRF_SoapySDR::writeSetting(const std::string &key, const std::string &v
         if (std::find(std::begin(xb200_validSettings), std::end(xb200_validSettings), value) != std::end(xb200_validSettings))
         {
             // --> Valid setting has arrived
+            _xb200Mode = value;
 
             // Get attached expansion device
             bladerf_xb _bladerf_xb_attached = bladerf_xb::BLADERF_XB_NONE;
@@ -657,6 +832,7 @@ void bladeRF_SoapySDR::writeSetting(const std::string &key, const std::string &v
         if (std::find(std::begin(sampling_mode_validSettings), std::end(sampling_mode_validSettings), value) != std::end(sampling_mode_validSettings))
         {
             // --> Valid setting has arrived
+            _samplingMode = value;
 
             // Set the sampling mode
             int ret = 0;
@@ -692,6 +868,7 @@ void bladeRF_SoapySDR::writeSetting(const std::string &key, const std::string &v
         if (std::find(std::begin(loopback_validSettings), std::end(loopback_validSettings), value) != std::end(loopback_validSettings))
         {
             // --> Valid setting has arrived
+            _loopbackMode = value;
 
             // Which loopback mode was selected?
             bladerf_loopback loopback = bladerf_loopback::BLADERF_LB_NONE;
